"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable-next-line: no-var-requires
const fs = require('fs');
const packageJson = require('../package.json');
const knexfile = require('../knexfile.js');
const axios_1 = __importDefault(require("axios"));
const body_parser_1 = require("body-parser");
const knex_1 = __importDefault(require("knex"));
const multer_1 = __importDefault(require("multer"));
const https_1 = __importDefault(require("https"));
const http_1 = __importDefault(require("http"));
const backtrace_service_1 = require("backtrace-service");
const express_1 = __importDefault(require("express"));
const ClangAnalyzer_1 = require("./application/adaptors/ClangAnalyzer");
// import { Parasoft } from './application/adaptors/Parasoft';
const send_1 = require("./application/send");
const utilities_1 = require("./utilities");
const Queue_1 = require("./application/Queue");
const Correlate_1 = require("./application/Correlate");
const utilities_2 = require("./utilities");
const log = backtrace_service_1.Logger.getLogger();
/**
 * Application name and default port required by service layer
 */
const name = 'sca';
const port = 6449;
const app = express_1.default();
/**
 * Create Backtrace Service
 */
const service = backtrace_service_1.BacktraceService.bootstrap(name, port, {}, app);
service.app.use(body_parser_1.json());
service.app.use(body_parser_1.urlencoded({ extended: true }));
/**
 * Multer instance for uploaded files.
 */
const storage = multer_1.default.memoryStorage();
const upload = multer_1.default({
    storage: storage,
    limits: {
        // fileSize: 1e+8, // removed for unlimited
        files: 1,
        fields: 10,
        parts: 11,
    }
});
/**
 * Set up database
 */
const dbConfig = knexfile[process.env["NODE_ENV"] || 'production'];
const knx = knex_1.default(dbConfig);
/**
 * Initialize storage directory
 * */
fs.mkdir(service.configuration.storage.path, { recursive: true }, (err) => {
    if (err) {
        throw new Error(`Error: tmp directory ${service.configuration.storage.path} cannot be created.`);
    }
});
/**
 * Set up global HTTP/HTTPS agents
 */
const agentOptions = ((_a = service.configuration.ingestion) === null || _a === void 0 ? void 0 : _a.concurrency) ?
    { maxSockets: service.configuration.ingestion.concurrency } : {};
exports.httpsAgent = new https_1.default.Agent(agentOptions);
exports.httpAgent = new http_1.default.Agent(agentOptions);
/**
 * Set up accepted formats
 * Removes formats that might have been introduced in the sca.conf but are not actually supported
 */
if (!service.configuration.formats || !Array.isArray(service.configuration.formats)) {
    throw new Error("Missing formats array in config.");
}
const availableFormats = service.configuration.formats.filter((format) => {
    return (
    // format === 'parasoft' ||
    format === 'clang-analyzer');
});
/**
 * Check for required config values (fail fast) and set defaults
 */
if (!service.configuration.tar) {
    throw new Error('Missing tar in config.');
}
if (availableFormats.length === 0) {
    throw new Error('No supported formats in config.');
}
const batchSize = ((_b = service.configuration.ingestion) === null || _b === void 0 ? void 0 : _b.batch) || 100;
log.info(`Submission will process in batches of ${batchSize} reports. (Configurable in sca.conf ingestion.batch)`);
/**
 * Submission
 */
service.app.post('/api/sca/submit/:format', upload.single('report'), async (req, res, next) => {
    const { format } = req.params;
    const { submitUrl } = req.body;
    // handle user error
    if (!submitUrl) {
        return res.send({ message: `request missing 'submitUrl' value` });
    }
    if (!req.file) {
        return res.send({ message: `request missing 'report' file value` });
    }
    const defectProjectUnivserseToken = utilities_1.getUniverseToken(submitUrl);
    if (!defectProjectUnivserseToken || !defectProjectUnivserseToken.token ||
        !defectProjectUnivserseToken.universe) { // abort submission if no token/universe
        return res.send({ message: `unable to parse 'submitUrl' for project token and universe` });
    }
    const { buffer } = req.file;
    // add attributes from querystring params
    const attrs = {};
    for (let k in req.query) {
        attrs[k] = req.query[k];
    }
    // Select adaptor for requested format
    let options = {
        data: buffer,
        config: service.configuration,
        logger: log,
    };
    let adaptor;
    switch (format) {
        // case 'parasoft': // disabled for now, not maintained.
        //   adaptor = new Parasoft({ ...options, data: buffer.toString('utf8')});
        //   break;
        case 'clang-analyzer':
            // clang-analyzer requires tar upload so NOT converting buffer to string here
            adaptor = new ClangAnalyzer_1.ClangAnalyzer(options);
            break;
        default:
            const unknownFormatError = new backtrace_service_1.ApiError(`Error: Format must be one of type: [${availableFormats}]`, 422);
            return next(unknownFormatError);
    }
    // let reports: IReportObject[];
    let batchIdentifier = 0;
    let successCount = 0;
    let failCount = 0;
    try {
        // Get a batch of reports, send them, delete them, get next batch.
        const generator = adaptor.reportsGenerator(batchSize);
        let generatorState = generator.next(); // initializes data and get first N reports
        while (!generatorState.done) {
            const reports = generatorState.value;
            const reportData = [];
            const r = reports.map((obj) => {
                var _a, _b, _c;
                // Extract defect data to store in local defect table
                const { attributes, classifiers, threads } = obj.report;
                reportData.push({
                    fingerprint: attributes._mod_fingerprint,
                    build: attributes.build,
                    errorMessage: attributes['error.message'],
                    classifiers,
                    lineNumber: (_a = threads) === null || _a === void 0 ? void 0 : _a.main.stack[0].line,
                    functionName: (_b = threads) === null || _b === void 0 ? void 0 : _b.main.stack[0].funcName,
                    fileName: (_c = threads) === null || _c === void 0 ? void 0 : _c.main.stack[0].library
                });
                // Add querystring attributes, convert to FormData
                obj.report.attributes = Object.assign(Object.assign({}, obj.report.attributes), attrs);
                return utilities_1.reportToFormData(obj, log);
            });
            // send: submits to coroner, stores defects in DB, returns a response object
            const { batchId, received, status } = await send_1.send(submitUrl, service.metrics, r, reportData, format, correlation, defectProjectUnivserseToken.universe, defectProjectUnivserseToken.token, ++batchIdentifier);
            if (status != 'ok') {
                log.error(`Batch #${batchId} failed to complete send. Received: ${received}. submitUrl: ${submitUrl}`);
                failCount += received;
            }
            else {
                log.info(`Batch #${batchId} completed sending. Received: ${received}. submitUrl: ${submitUrl}`);
                successCount += received;
            }
            // now loop can continue - get next N reports.
            generatorState = generator.next();
        }
    }
    catch (e) {
        next(e);
    }
    finally {
        adaptor.close();
        res.send({ "status": "ok", successCount, failCount });
    }
});
/**
 * [DEPRECATED] Submission
 */
service.app.post('/api/sca/submit/:universe/:token/:format', async (req, res, next) => {
    res.send({ message: "This route is deprecated. For defect submission, please use /api/sca/submit/:format" });
});
/**
 * Correlation Request
 * Requests are pushed into a queue that will run correlation
 */
const correlation = new Correlate_1.Correlate(service.configuration, knx, service.metrics);
const queue = new Queue_1.Queue(correlation.correlate);
// PUT triggers correlation processing
const requiredParameters = ["crashProject", "crashProjectToken", "defectProject", "defectProjectToken", "fingerprint", "queryUrl"];
service.app.put('/api/sca/correlate/:universe', (req, res, next) => {
    const { universe } = req.params;
    for (const idx in requiredParameters) {
        if (!req.body[requiredParameters[idx]]) {
            return res.send({ message: `missing required param ${requiredParameters[idx]}` });
        }
    }
    const { crashProject, crashProjectToken, defectProject, defectProjectToken, fingerprint, objectData, ignoreFiles, ignoreFunctions, limitFiles, limitFunctions, classifierMatch, queryUrl, } = req.body;
    // Validate params
    const options = {
        ignoreFiles: Array.isArray(ignoreFiles) ? ignoreFiles : undefined,
        ignoreFunctions: Array.isArray(ignoreFunctions) ? ignoreFunctions : undefined,
        limitFiles: Number.isInteger(limitFiles) ? limitFiles : undefined,
        limitFunctions: Number.isInteger(limitFunctions) ? limitFunctions : undefined,
        classifierMatch: [true, false].includes(classifierMatch) ? classifierMatch : false,
    };
    // Submit for correlation processing
    queue.enqueue({
        universe,
        crashProject,
        crashProjectToken,
        defectProject,
        defectProjectToken,
        fingerprint,
        options,
        startTime: process.hrtime.bigint(),
        crashObject: objectData,
        queryUrl,
    });
    const message = `Submitted fingerprint ${fingerprint} of universe/crashProject ${universe}/${crashProject} for correlation.`;
    log.info(message);
    res.send({ message });
    service.metrics.addMetric(`correlate-put-enqueue`, 1);
});
//  GET returns correlated defects for a given fingerprint
service.app.get('/api/sca/correlate/:universe/:crashProject/:fingerprint', service.identityManager.authenticateRequest(), async (req, res, next) => {
    const authRequest = req;
    if (!authRequest ||
        !authRequest.coronerAuth ||
        !authRequest.coronerAuth.cfg ||
        !authRequest.coronerAuth.token) {
        service.metrics.addMetric(`correlate-get-auth-failure`, 1);
        return res.json({ status: 'unauthorized' });
    }
    const { universe, fingerprint, crashProject } = req.params;
    const { build, fileName, functionName } = req.query;
    try {
        const defects = await correlation.getCorrelatedDefects(universe, crashProject, fingerprint, { build, fileName, functionName });
        res.send({
            fingerprint,
            universe,
            crashProject,
            defects,
            build,
            fileName,
            functionName,
        });
        service.metrics.addMetric(`correlate-get-success`, 1);
    }
    catch (e) {
        service.metrics.addMetric(`correlate-get-error`, 1);
        next(e);
    }
});
/**
 * Backfill Correlation Request
 *  Trigger correlation processing for every fingerprint in project
 */
const requiredBackfillParameters = ["crashProject", "crashProjectToken", "defectProject", "defectProjectToken", "queryUrl"];
service.app.put('/api/sca/backfill-correlation/:universe', async (req, res, next) => {
    const { universe } = req.params;
    for (const idx in requiredBackfillParameters) {
        if (!req.body[requiredBackfillParameters[idx]]) {
            return res.send({ message: `missing required param ${requiredBackfillParameters[idx]}` });
        }
    }
    const { crashProject, crashProjectToken, defectProject, defectProjectToken, objectData, ignoreFiles, ignoreFunctions, limitFiles, limitFunctions, classifierMatch, queryUrl, } = req.body;
    // Validate params
    const options = {
        ignoreFiles: Array.isArray(ignoreFiles) ? ignoreFiles : undefined,
        ignoreFunctions: Array.isArray(ignoreFunctions) ? ignoreFunctions : undefined,
        limitFiles: Number.isInteger(limitFiles) ? limitFiles : undefined,
        limitFunctions: Number.isInteger(limitFunctions) ? limitFunctions : undefined,
        classifierMatch: [true, false].includes(classifierMatch) ? classifierMatch : false,
    };
    // Get fingerprints of this project
    const date = new Date();
    date.setDate(date.getDate() - 365 / 4);
    const pastQuarter = Math.round(date.getTime() / 1000);
    const queryForCrashes = utilities_2.queryForCrashesSince(pastQuarter.toString(10));
    const url = `${queryUrl}?token=${crashProjectToken}&project=${crashProject}`;
    let results;
    try {
        results = await axios_1.default.post(url, queryForCrashes);
    }
    catch (e) {
        log.error(e);
        return res.send({ message: `unable to fetch crashes for ${universe}/${crashProject}` });
    }
    const crashObjectsRaw = results.data;
    if (crashObjectsRaw.error) {
        const message = `Error: "${crashObjectsRaw.error.message}"`;
        log.error(message);
        return res.send({ message });
        ;
    }
    const crashObjectsUnpacked = Correlate_1.Correlate.unpack(crashObjectsRaw.response);
    // Submit each crash fingerprint for correlation processing
    let submittedCount = 0;
    Object.keys(crashObjectsUnpacked).forEach(fingerprint => {
        queue.enqueue({
            universe,
            crashProject,
            crashProjectToken,
            defectProject,
            defectProjectToken,
            fingerprint,
            options,
            startTime: process.hrtime.bigint(),
            crashObject: objectData,
            queryUrl,
        });
        submittedCount++;
        log.debug(`Submitted fingerprint ${fingerprint} of ${universe}/${crashProject} for correlation.`);
    });
    const message = `Submitted ${submittedCount} fingerprints from ${universe}/${crashProject} for backfill correlation.`;
    log.info(message);
    res.send({ message });
    service.metrics.addMetric(`correlate-backfill`, 1);
});
/**
 * Service Version
 */
service.app.get('/api/sca/version', (req, res) => {
    res.json({
        response: { status: 'ok' },
        _: {
            version: packageJson.version,
        },
    });
});
service.app.get('*', (req, res) => {
    res.status(404).send();
});
/**
 * Start
 */
service.start(service.configuration.clusterSupport);
exports.default = service;
//# sourceMappingURL=main.js.map